import numpy as np
from numba import njit, cuda
from numba.cuda import random
import math
from numba import float64, int32

@cuda.jit
def symplectic_map_common(x, px, step_values, noise_array, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius):
    i = cuda.threadIdx.x
    j = cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x

    action = cuda.shared.array(shape=(512), dtype=float64)
    rot_angle = cuda.shared.array(shape=(512), dtype=float64)
    temp1 = cuda.shared.array(shape=(512), dtype=float64)
    temp2 = cuda.shared.array(shape=(512), dtype=float64)
    l_x = cuda.shared.array(shape=(512), dtype=float64)
    l_px = cuda.shared.array(shape=(512), dtype=float64)
    l_step = cuda.shared.array(shape=(512), dtype=int32)
    
    if j < x.shape[0]:
        l_x[i] = x[j]
        l_px[i] = px[j]
        l_step[i] = step_values[j]
        for k in range(noise_array.shape[0]):
            action[i] = (l_x[i] * l_x[i] + l_px[i] * l_px[i]) * 0.5
            rot_angle[i] = omega_0 + (omega_1 + action[i]) + (0.5 * omega_2 * action[i] * action[i])

            if (l_x[i] == 0.0 and px[i] == 0.0) or action[i] >= action_radius:
                l_x[i] = 0.0
                l_px[i] = 0.0
                break

            temp1[i] = l_x[i]
            temp2[i] = (
                l_px[i] + epsilon * noise_array[k] * (l_x[i] ** beta)
                * math.exp(-((x_star / (delta + abs(l_x[i]))) ** alpha))
            )
            l_x[i] = math.cos(rot_angle[i]) * temp1[i] + \
                math.sin(rot_angle[i]) * temp2[i]
            l_px[i] = -math.sin(rot_angle[i]) * temp1[i] + \
                math.cos(rot_angle[i]) * temp2[i]

            l_step[i] += 1
        x[j] = l_x[i]
        px[j] = l_px[i]
        step_values[j] = l_step[i]


@cuda.jit
def symplectic_map_personal(x, px, step_values, n_iterations, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius, rng_states, gamma):
    i = cuda.threadIdx.x
    j = cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x

    action = cuda.shared.array(shape=(512), dtype=float64)
    rot_angle = cuda.shared.array(shape=(512), dtype=float64)
    temp1 = cuda.shared.array(shape=(512), dtype=float64)
    temp2 = cuda.shared.array(shape=(512), dtype=float64)
    noise = cuda.shared.array(shape=(512), dtype=float64)
    l_x = cuda.shared.array(shape=(512), dtype=float64)
    l_px = cuda.shared.array(shape=(512), dtype=float64)
    l_step = cuda.shared.array(shape=(512), dtype=int32)

    noise[i] = random.xoroshiro128p_normal_float64(rng_states, j)

    if j < x.shape[0]:
        l_x[i] = x[j]
        l_px[i] = px[j]
        l_step[i] = step_values[j]
        for k in range(n_iterations):
            action[i] = (l_x[i] * l_x[i] + l_px[i] * l_px[i]) * 0.5
            rot_angle[i] = omega_0 + (omega_1 + action[i]) + \
                (0.5 * omega_2 * action[i] * action[i])

            if (l_x[i] == 0.0 and l_px[i] == 0.0) or action[i] >= action_radius:
                l_x[i] = 0.0
                l_px[i] = 0.0
                break
            
            temp1[i] = l_x[i]
            temp2[i] = (
                l_px[i] + epsilon * noise[i] * (l_x[i] ** beta)
                * math.exp(-((x_star / (delta + abs(l_x[i]))) ** alpha))
            )
            l_x[i] = math.cos(rot_angle[i]) * temp1[i] + \
                math.sin(rot_angle[i]) * temp2[i]
            l_px[i] = -math.sin(rot_angle[i]) * temp1[i] + \
                math.cos(rot_angle[i]) * temp2[i]

            l_step[i] += 1

            noise[i] = random.xoroshiro128p_normal_float64(rng_states, j) + gamma * noise[i]
        x[j] = l_x[i]
        px[j] = l_px[i]
        step_values[j] = l_step[i]



import numpy as np
from numba import jit, njit, prange


@njit
def make_correlated_noise(n_elements, gamma=0.0):
    """Make an array of correlated noise
    
    Parameters
    ----------
    n_elements : unsigned int
        number of elements
    gamma : float, optional
        correlation coefficient, by default 0.0
    
    Returns
    -------
    ndarray
        the noise array
    """    
    noise = np.random.normal(0.0, 1.0, n_elements)
    if gamma != 0.0:
        for i in range(1, n_elements):
            noise[i] += gamma * noise[i - 1]
    return noise


@njit
def iterate(x, px, noise, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius, start):
    """internal iteration method for symplectic map
    
    Parameters
    ----------
    x : float
        x0
    px : float
        px0
    noise : ndarray
        array of noise values
    epsilon : float
        epsilon value
    alpha : float
        alpha exponential
    beta : float
        beta exponential
    x_star : float
        nek coefficient
    delta : float
        delta coefficient for fix
    omega_0 : float
        omega 0
    omega_1 : float
        omega 1
    omega_2 : float
        omega 2
    action_radius : float
        barrier position in action value
    start : unsigned int
        starting iteration value
    
    Returns
    -------
    (float, float, unsigned int)
        (x, px, iterations)
    """    
    for i in range(len(noise)):
        action = (x * x + px * px) * 0.5
        rot_angle = omega_0 + (omega_1 * action) + (0.5 * omega_2 * action * action)
        # rot_angle = np.random.rand() * np.pi * 2
        
        if (x == 0 and px == 0) or action >= action_radius:
            return 0.0, 0.0, i + start
        
        temp1 = x
        temp2 = (px + epsilon * noise[i] * np.power(x, beta) *
                 np.exp(-np.power((x_star / (delta + np.abs(x))), alpha)))
        x = np.cos(rot_angle) * temp1 + np.sin(rot_angle) * temp2
        px = -np.sin(rot_angle) * temp1 + np.cos(rot_angle) * temp2

    return x, px, i + start


@njit(parallel=True)
def symplectic_map_personal(x, px, step_values, n_iterations, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius, gamma=0.0):
    """computation for personal noise symplectic map
    
    Parameters
    ----------
    x : ndarray
        x initial condition
    px : ndarray
        px initial condition
    step_values : ndarray
        iterations already performed
    n_iterations : unsigned int
        number of iterations to perform
    epsilon : float
        epsilon value
    alpha : float
        alpha exponential
    beta : float
        beta exponential
    x_star : float
        nek coefficient
    delta : float
        nek coefficient
    omega_0 : float
        ipse dixit
    omega_1 : float
        ipse dixit
    omega_2 : float
        ipse dixit
    action_radius : float
        barrier radius in action value!
    gamma : float, optional
        correlation coefficient, by default 0.0
    
    Returns
    -------
    (ndarray, ndarray, ndarray)
        x, px, step_values
    """    
    for i in prange(len(x)):
        personal_noise = make_correlated_noise(n_iterations, gamma)
        x[i], px[i], step_values[i] = iterate(x[i], px[i], personal_noise, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius, step_values[i])
    return x, px, step_values 



@njit(parallel=True)
def symplectic_map_common(x, px, step_values, noise_array, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius):
    """computation for personal noise symplectic map

    Parameters
    ----------
    x : ndarray
        x initial condition
    px : ndarray
        px initial condition
    step_values : ndarray
        iterations already performed
    noise_array : ndarray
        noise array for the whole group
    epsilon : float
        epsilon value
    alpha : float
        alpha exponential
    beta : float
        beta exponential
    x_star : float
        nek coefficient
    delta : float
        nek coefficient
    omega_0 : float
        ipse dixit
    omega_1 : float
        ipse dixit
    omega_2 : float
        ipse dixit
    action_radius : float
        barrier radius in action value!
    gamma : float, optional
        correlation coefficient, by default 0.0

    Returns
    -------
    (ndarray, ndarray, ndarray)
        x, px, step_values
    """
    for i in prange(len(x)):
        x[i], px[i], step_values[i] = iterate(x[i], px[i], noise_array, epsilon, alpha, beta, x_star, delta, omega_0, omega_1, omega_2, action_radius, step_values[i])
    return x, px, step_values 



from setuptools import setup
import sys
import setuptools
import glob

__version__ = '0.0.4'

setup(
    name='symplectic_map',
    version=__version__,
    author='Carlo Emilio Montanari',
    author_email='carlidel95@gmail.com',
    url='https://github.com/carlidel/c_symplectic_map',
    description='A Numba implementation of a Symplectic Map with nice python bindings',
    packages=['symplectic_map'],
    install_requires=['numba', 'numpy', 'scipy', 'tqdm'],
    setup_requires=['numba', 'numpy', 'scipy', 'tqdm'],
    license='MIT',
)



import numpy as np
from tqdm import tqdm
import scipy as sc
import scipy.integrate as integrate
from numba import cuda
import numba.cuda.random as random

from . import cpu_symplectic_map as cpu
from . import gpu_symplectic_map as gpu


def make_correlated_noise(n_elements, gamma=0.0):
    return cpu.make_correlated_noise(n_elements, gamma)


class symplectic_map(object):
    def __init__(self):
        pass
    
    def reset(self):
        pass

    def compute_common_noise(self):
        pass
    
    def compute_personal_noise(self):
        pass
    
    def get_data(self):
        """Get data from engine.
        
        Returns
        -------
        tuple(ndarray, ndarray, ndarray)
            tuple with x, p, and number of iterations before loss data
        """
        return self.x, self.p, self.times

    def get_filtered_data(self):
        """Get filtered data from engine.
        
        Returns
        -------
        tuple(ndarray, ndarray, ndarray)
            tuple with x, p, and number of iterations before loss data
        """
        t = self.times
        return (self.x)[t == self.iterations], (self.p)[t == self.iterations], t[t == self.iterations]

    def get_action(self):
        """Get action data from engine
        
        Returns
        -------
        ndarray
            action array data
        """
        return (self.x * self.x + self.p * self.p) * 0.5

    def get_filtered_action(self):
        """Get filtered action data from engine (i.e. no zero values of lost particles)
        
        Returns
        -------
        ndarray
            filtered action array data
        """
        action = self.get_action()
        return action[action > 0]

    def get_times(self):
        """Get times from engine
        
        Returns
        -------
        ndarray
            times array
        """
        return self.times

    def get_filtered_times(self):
        """Get only loss times from engine (i.e. only loss particles)
        
        Returns
        -------
        ndarray
            filtered times array
        """
        times = self.get_times()
        return times[times < self.iterations]

    def get_survival_quota(self):
        """Get time evolution of number of survived particles
        
        Returns
        -------
        ndarray
            time evolution of survived particles
        """
        t = np.array(self.get_times())
        max_t = np.amax(t)
        quota = np.empty(max_t)
        for i in range(max_t):
            quota[i] = np.count_nonzero(t > i)
        return quota

    def get_lost_particles(self):
        """Get time evolution of lost particles
        
        Returns
        -------
        ndarray
            time evolution of number of lost particles
        """
        quota = self.get_survival_quota()
        return self.N - quota

    def current_binning(self, bin_size):
        """Execute current binning and computation
        
        Parameters
        ----------
        bin_size : int
            size of the binning to consider for current computation
        
        Returns
        -------
        tuple(ndarray, ndarray)
            array with corresponding sampling time (middle point), current value computed.
        """
        survival_quota = self.get_survival_quota()
        points = [i for i in range(0, len(survival_quota), bin_size)]
        if survival_quota % bin_size == 0:
            points.append(len(survival_quota) - 1)
        t_middle = [(points[i + 1] + points[i]) *
                    0.5 for i in range(len(points) - 1)]
        currents = [(survival_quota[points[i]] - survival_quota[points[i+1]]
                     ) / bin_size for i in range(len(points) - 1)]
        return np.array(t_middle), np.array(currents)

    @staticmethod
    def generate_instance(omega_0, omega_1, omega_2, epsilon, x_star, delta, alpha, beta, barrier_radius, x_0, p_0, cuda_device=None):
        if cuda_device == None:
            cuda_device = cuda.is_available()
        if cuda_device:
            return symplectic_map_gpu(omega_0, omega_1, omega_2, epsilon, x_star, delta, alpha, beta, barrier_radius, x_0, p_0)
        else:
            return symplectic_map_cpu(omega_0, omega_1, omega_2, epsilon, x_star, delta, alpha, beta, barrier_radius, x_0, p_0)

class symplectic_map_cpu(symplectic_map):
    def __init__(self, omega_0, omega_1, omega_2, epsilon, x_star, delta, alpha, beta, barrier_radius, x_0, p_0):
        """Init symplectic map object!
        
        Parameters
        ----------
        object : self
            self
        omega_0 : float
            Omega 0 frequency
        omega_1 : float
            Omega 1 frequency
        omega_2 : float
            Omega 2 frequency
        epsilon : float
            Noise coefficient
        x_star : float
            X star Nekhoroshev coefficient
        delta : float
            value to avoid singularities for zero action values
        alpha : float
            alpha exponential Nekhoroshev coefficient
        beta : float
            beta polynomial exponent
        barrier_radius : float
            barrier position (x coordinates!)
        x_0 : ndarray
            1D array of x initial positions
        p_0 : ndarray
            1D array of p initial values
        """
        self.omega_0 = omega_0
        self.omega_1 = omega_1
        self.omega_2 = omega_2
        self.epsilon = epsilon
        self.x_star = x_star
        self.delta = delta
        self.alpha = alpha
        self.beta = beta
        self.barrier_radius = barrier_radius
        self.action_radius = barrier_radius ** 2 * 0.5
        self.x_0 = x_0.copy()
        self.p_0 = p_0.copy()
        self.N = len(x_0)

        self.iterations = 0

        self.x = x_0.copy()
        self.p = p_0.copy()
        self.times = np.zeros(len(self.x))

    def reset(self):
        """Reset the engine to initial conditions
        """
        self.iterations = 0
        self.x = self.x_0.copy()
        self.p = self.p_0.copy()
        self.times = np.zeros(len(self.x))

    def compute_common_noise(self, noise_array):
        """Execute iterations with given noise array, common for all particles.
        
        Parameters
        ----------
        noise_array : ndarray
            noise array to use for computation
        """
        self.iterations += len(noise_array)
        self.x, self.p, self.times = cpu.symplectic_map_common(
            self.x, self.p, self.times, noise_array, self.epsilon, self.alpha, self.beta, self.x_star, self.delta, self.omega_0, self.omega_1, self.omega_2, self.action_radius
        )

    def compute_personal_noise(self, n_iterations, gamma=0.0):
        """Execute iterations with correlated noise with different realization for every single particle.
        
        Parameters
        ----------
        n_iterations : unsigned int
            number of iterations to perform
        gamma : float, optional
            correlation coefficient (between 0 and 1!), by default 0.0
        """
        self.iterations += n_iterations
        self.x, self.p, self.times = cpu.symplectic_map_personal(
            self.x, self.p, self.times, n_iterations, self.epsilon, self.alpha, self.beta, self.x_star, self.delta, self.omega_0, self.omega_1, self.omega_2, self.action_radius, gamma
        )


class symplectic_map_gpu(symplectic_map):
    def __init__(self, omega_0, omega_1, omega_2, epsilon, x_star, delta, alpha, beta, barrier_radius, x_0, p_0):
        """Init symplectic map object!
        
        Parameters
        ----------
        object : self
            self
        omega_0 : float
            Omega 0 frequency
        omega_1 : float
            Omega 1 frequency
        omega_2 : float
            Omega 2 frequency
        epsilon : float
            Noise coefficient
        x_star : float
            X star Nekhoroshev coefficient
        delta : float
            value to avoid singularities for zero action values
        alpha : float
            alpha exponential Nekhoroshev coefficient
        beta : float
            beta polynomial exponent
        barrier_radius : float
            barrier position (x coordinates!)
        x_0 : ndarray
            1D array of x initial positions
        p_0 : ndarray
            1D array of p initial values
        """
        self.omega_0 = omega_0
        self.omega_1 = omega_1
        self.omega_2 = omega_2
        self.epsilon = epsilon
        self.x_star = x_star
        self.delta = delta
        self.alpha = alpha
        self.beta = beta
        self.barrier_radius = barrier_radius
        self.action_radius = barrier_radius ** 2 * 0.5
        self.x_0 = x_0.copy()
        self.p_0 = p_0.copy()
        self.N = len(x_0)

        self.iterations = 0

        self.x = x_0.copy()
        self.p = p_0.copy()
        self.times = np.zeros(len(self.x))

        # Load to GPU
        self.d_x = cuda.to_device(self.x)
        self.d_p = cuda.to_device(self.p)
        self.d_times = cuda.to_device(self.times)

    def reset(self):
        """Reset the engine to initial conditions
        """
        self.iterations = 0
        self.x = self.x_0.copy()
        self.p = self.p_0.copy()
        self.times = np.zeros(len(self.x))
        self.d_x = cuda.to_device(self.x)
        self.d_p = cuda.to_device(self.p)
        self.d_times = cuda.to_device(self.times)

    def compute_common_noise(self, noise_array):
        """Execute iterations with given noise array, common for all particles.
        
        Parameters
        ----------
        noise_array : ndarray
            noise array to use for computation
        """
        threads_per_block = 512
        blocks_per_grid = self.x.size // 512 + 1

        self.iterations += len(noise_array)
        d_noise_array = cuda.to_device(noise_array)
        gpu.symplectic_map_common[blocks_per_grid, threads_per_block](
            self.d_x, self.d_p, self.d_times, d_noise_array, self.epsilon, self.alpha, self.beta, self.x_star, self.delta, self.omega_0, self.omega_1, self.omega_2, self.action_radius
        )
        self.d_x.copy_to_host(self.x)
        self.d_p.copy_to_host(self.p)
        self.d_times.copy_to_host(self.times)

    def compute_personal_noise(self, n_iterations, gamma=0.0):
        """Execute iterations with correlated noise with different realization for every single particle.
        
        Parameters
        ----------
        n_iterations : unsigned int
            number of iterations to perform
        gamma : float, optional
            correlation coefficient (between 0 and 1!), by default 0.0
        """
        threads_per_block = 512
        blocks_per_grid = self.x.size // 512 + 1

        self.iterations += n_iterations
        rng_states = cuda.random.create_xoroshiro128p_states(
            threads_per_block * blocks_per_grid, seed=1)
        gpu.symplectic_map_personal[blocks_per_grid, threads_per_block](
            self.d_x, self.d_p, self.d_times, n_iterations, self.epsilon, self.alpha, self.beta, self.x_star, self.delta, self.omega_0, self.omega_1, self.omega_2, self.action_radius, rng_states, gamma
        )
        self.d_x.copy_to_host(self.x)
        self.d_p.copy_to_host(self.p)
        self.d_times.copy_to_host(self.times)